--
-- FS25 - FieldExporter
--
-- This mod exports field information to an XML file
--

FieldExporter = {}
local modDescFile = loadXMLFile("modDesc", g_currentModDirectory.."modDesc.xml")
FieldExporter.title = getXMLString(modDescFile, "modDesc.title.en")
FieldExporter.author = getXMLString(modDescFile, "modDesc.author")
FieldExporter.version = getXMLString(modDescFile, "modDesc.version")
FieldExporter.copyright = "Copyright © 2025 Sippler. All rights reserved."
delete(modDescFile)

-- Safe property access helper
function FieldExporter:safeGet(obj, key, default)
    if obj == nil then return default end
    
    -- Direct access
    local success, result = pcall(function() return obj[key] end)
    if success and result ~= nil then
        return result
    end
    
    -- Access via getter method
    if type(obj) == "table" then
        local func = obj["get" .. key:sub(1,1):upper() .. key:sub(2)]
        if type(func) == "function" then
            success, result = pcall(function() return func(obj) end)
            if success and result ~= nil then
                return result
            end
        end
    end
    
    return default
end

-- Safe number conversion
function FieldExporter:safeGetNumber(field, key, default)
    local value = self:safeGet(field, key, default)
    
    if type(value) == "number" then
        return value
    elseif type(value) == "string" then
        local num = tonumber(value)
        if num ~= nil then
            return num
        end
    end
    
    return default
end

-- Calculate field area
function FieldExporter:calculateFieldArea(field)
    local possibleAreaNames = {
        "areaHa", "fieldAreaHa", "size", "fieldArea", "area", "fieldSize", 
        "totalFieldArea", "hectares", "ha", "totalArea"
    }
    
    for _, areaName in ipairs(possibleAreaNames) do
        local value = self:safeGetNumber(field, areaName, 0)
        if value > 0 then
            -- Convert to hectares if needed
            if value > 1000 then  -- Probably in square meters
                return value / 10000  -- Convert to hectares (1 ha = 10000 m²)
            else
                return value  -- Already in hectares
            end
        end
    end
    
    -- Default value
    return 0.01
end

-- Mod initialization
function FieldExporter:loadMap(savegame)
    if g_currentMission:getIsClient() then
        FieldExporter.events = {}
        PlayerInputComponent.registerGlobalPlayerActionEvents = Utils.appendedFunction(PlayerInputComponent.registerGlobalPlayerActionEvents, FieldExporter.registerActionEvents)
        PlayerInputComponent.unregisterActionEvents = Utils.appendedFunction(PlayerInputComponent.unregisterActionEvents, FieldExporter.unregisterActionEvents)
        Enterable.onRegisterActionEvents = Utils.appendedFunction(Enterable.onRegisterActionEvents, FieldExporter.registerActionEvents)
    end
    
    -- Output mod info
    print(self.title .. " : v" .. self.version .. " by " .. self.author .. " activated")
end

function FieldExporter:unregisterActionEvents()
    if self.player and self.player.isClient then
        g_inputBinding:removeActionEvent(FieldExporter.events[1])
        FieldExporter.events = {}
    end
end

function FieldExporter:registerActionEvents()
    local isClient = self.isClient
    if isClient == nil then
        isClient = self.player ~= nil and self.player.isClient
    end
    
    if isClient then
        FieldExporter.events = {}
        local valid, eventId = g_inputBinding:registerActionEvent(InputAction.FIELD_EXPORTER_EXPORT, InputBinding.NO_EVENT_TARGET, FieldExporter.exportFieldsHandler, false, true, false, true)
        
        if valid then
            table.insert(FieldExporter.events, eventId)
            g_inputBinding:setActionEventText(eventId, g_i18n:getText("input_FIELD_EXPORTER_EXPORT"))
            g_inputBinding:setActionEventTextVisibility(eventId, true)
            g_inputBinding:setActionEventTextPriority(eventId, GS_PRIO_VERYLOW)
        end
    end
end

function FieldExporter:exportFieldsHandler(actionName, keyStatus, arg3, arg4, arg5)
    if actionName == "FIELD_EXPORTER_EXPORT" then
        FieldExporter:exportFields()
    end
end

-- Main export function
function FieldExporter:exportFields()
    if g_currentMission == nil then 
        print("FieldExporter: Current mission is nil")
        return 
    end
    
    -- Create export directory
    local modSettingsDir = getUserProfileAppPath() .. "modSettings/"
    createFolder(modSettingsDir)
    
    local exportDir = modSettingsDir .. "FS25_FieldExporter/"
    createFolder(exportDir)
    
    -- Determine map name
    local mapName = "unknown"
    if g_currentMission.missionInfo ~= nil and g_currentMission.missionInfo.mapTitle ~= nil then
        mapName = g_currentMission.missionInfo.mapTitle
    elseif g_currentMission.missionInfo ~= nil and g_currentMission.missionInfo.map ~= nil and g_currentMission.missionInfo.map.title ~= nil then
        mapName = g_currentMission.missionInfo.map.title
    end
    
    -- Remove invalid characters
    mapName = mapName:gsub("[%c%p%s]", "_")
    
    -- Create XML file
    local filename = string.format("fields_%s.xml", mapName)
    local xmlFilename = exportDir .. filename
    local xmlFile = createXMLFile("FieldExport", xmlFilename, "fields")
    
    -- Add copyright information to XML
    setXMLString(xmlFile, "fields#copyright", FieldExporter.copyright)
    setXMLString(xmlFile, "fields#licenseNote", "This export file is for personal use only and may not be redistributed.")
    setXMLString(xmlFile, "fields#version", FieldExporter.version)
    
    -- Collect all fields
    local fields = {}
    local fieldCount = 0
    
    -- Safe field collection function
    local function safeCollectFields(source, name)
        if source == nil or type(source) ~= "table" then
            return 0
        end
        
        local count = 0
        
        for _, field in pairs(source) do
            if field ~= nil and type(field) == "table" then
                count = count + 1
                table.insert(fields, field)
            end
        end
        
        if count > 0 then
            print("FieldExporter: " .. count .. " fields added from " .. name)
        end
        
        return count
    end
    
    -- Collect fields from various sources
    if g_fieldManager ~= nil then
        local possibleFieldCollections = {
            "fields", "numericIndexedFields", "fieldNumUpdateTasks", 
            "cpCustomFields", "customFields", "allFields", "farmFields", "fieldMap"
        }
        
        for _, collectionName in ipairs(possibleFieldCollections) do
            local collection = self:safeGet(g_fieldManager, collectionName, nil)
            fieldCount = fieldCount + safeCollectFields(collection, "g_fieldManager." .. collectionName)
        end
        
        -- Additional field-related collections
        for k, v in pairs(g_fieldManager) do
            if type(k) == "string" and type(v) == "table" and 
               (k:find("field") or k:find("Field")) then
                if not (k == "fields" or k == "numericIndexedFields" or 
                       k == "fieldNumUpdateTasks" or k == "cpCustomFields" or
                       k == "customFields" or k == "allFields" or
                       k == "farmFields" or k == "fieldMap") then
                    fieldCount = fieldCount + safeCollectFields(v, "g_fieldManager." .. k)
                end
            end
        end
    end
    
    -- Additional field sources
    fieldCount = fieldCount + safeCollectFields(self:safeGet(g_currentMission, "fields", nil), "g_currentMission.fields")
    fieldCount = fieldCount + safeCollectFields(g_fields, "g_fields")
    
    print(string.format("FieldExporter: Total %d fields found", #fields))
    
    -- Determine unique fields
    local fieldDataList = {}
    
    for i, field in ipairs(fields) do
        -- Extract field ID
        local fieldId = 0
        
        local possibleIdNames = {"id", "fieldId", "fieldID", "index", "number", "fieldNum", "fieldNumber"}
        for _, idName in ipairs(possibleIdNames) do
            local value = self:safeGetNumber(field, idName, 0)
            if value > 0 then
                fieldId = value
                break
            end
        end
        
        -- Use index if no ID found
        if fieldId <= 0 then
            fieldId = i
        end
        
        -- Calculate area
        local fieldArea = self:calculateFieldArea(field)
        
        -- Round to 2 decimal places
        fieldArea = math.floor(fieldArea * 100 + 0.5) / 100
        
        -- Safely extract farm ID
        local farmId = 0
        
        -- Direct farmId property
        farmId = self:safeGetNumber(field, "farmId", 0)
        
        -- Alternative owner property
        if farmId == 0 then
            farmId = self:safeGetNumber(field, "owner", 0) or 
                     self:safeGetNumber(field, "ownerId", 0)
        end
        
        -- Try getter functions
        if farmId == 0 and type(field.getOwner) == "function" then
            local success, result = pcall(function() return field:getOwner() end)
            if success and type(result) == "number" then
                farmId = result
            end
        end
        
        if farmId == 0 and type(field.getFarmId) == "function" then
            local success, result = pcall(function() return field:getFarmId() end)
            if success and type(result) == "number" then
                farmId = result
            end
        end
        
        -- Add to list
        table.insert(fieldDataList, {
            id = fieldId,
            area = fieldArea,
            farmId = farmId
        })
    end
    
    -- Remove duplicates
    local uniqueFields = {}
    local uniqueFieldsByID = {}
    
    for _, fieldData in ipairs(fieldDataList) do
        if not uniqueFieldsByID[fieldData.id] then
            uniqueFieldsByID[fieldData.id] = true
            table.insert(uniqueFields, fieldData)
        end
    end
    
    -- Sort by ID
    table.sort(uniqueFields, function(a, b)
        return a.id < b.id
    end)
    
    print(string.format("FieldExporter: %d unique fields after removing duplicates", #uniqueFields))
    
    -- Add map information
    setXMLString(xmlFile, "fields#mapName", mapName)
    setXMLInt(xmlFile, "fields#totalFields", #uniqueFields)
    
    -- Write field data to XML
    for i, fieldData in ipairs(uniqueFields) do
        local fieldKey = string.format("fields.field(%d)", i-1)
        
        -- Safely save XML properties
        local function safeSaveXML(tag, value, formatter)
            local success, error = pcall(function()
                local formattedValue = value
                if formatter then 
                    formattedValue = formatter(value)
                end
                
                if type(formattedValue) == "number" then
                    setXMLFloat(xmlFile, tag, formattedValue)
                else
                    setXMLString(xmlFile, tag, tostring(formattedValue))
                end
            end)
            
            if not success then
                print("FieldExporter: Warning - Could not save " .. tag .. ": " .. error)
            end
        end
        
        -- Save field properties
        safeSaveXML(fieldKey .. "#id", fieldData.id)
        safeSaveXML(fieldKey .. "#sizeHa", fieldData.area, function(v) return string.format("%.2f", v) end)
        safeSaveXML(fieldKey .. "#farmId", fieldData.farmId)
        
        -- Add farm name if available
        if fieldData.farmId > 0 then
            local success, farmName = pcall(function()
                if g_farmManager ~= nil then
                    local farm = g_farmManager:getFarmById(fieldData.farmId)
                    if farm ~= nil and farm.name ~= nil then
                        return farm.name
                    end
                end
                return nil
            end)
            
            if success and farmName ~= nil then
                safeSaveXML(fieldKey .. "#farmName", farmName)
            end
        end
    end
    
    -- Save and close XML file
    saveXMLFile(xmlFile)
    delete(xmlFile)
    
    -- Show notification
    if g_currentMission ~= nil then
        pcall(function()
            local message = string.format(g_i18n:getText("notification_fieldsExported"), #uniqueFields, "modSettings/FS25_FieldExporter/" .. filename)
            
            -- Standard notification
            if g_currentMission.addIngameNotification ~= nil then
                g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_INFO, message)
                g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_INFO, g_i18n:getText("copyright_notice"))
            end
            
            -- Direct console output
            print("FieldExporter: " .. message)
        end)
    end
    
    print("FieldExporter: " .. #uniqueFields .. " fields exported to " .. xmlFilename)
    return true
end

addModEventListener(FieldExporter)